import os
import shutil
import urllib.parse
import pathlib

import panflute as pf

from ...utils.logging_helper import TracingLogger

from ...utils.oss_helper import OssHelper

from ...utils.html_helper import get_html_src,sub_html_src

from ...utils.panflute_helper import decode_src_url

from ...utils.hash_helper import get_text_file_hash,get_raw_file_hash



r"""A pandoc filter that mainly for converting `markdown` to `markdown`.
Move local pictures to a new place. Replace the original `src` with the new one.
The new src will only remain the last directory name and the file name.
A root flag '/' will be added to the new src's header.

NOTE:
    The following environment variables should be given in advance:
        $Env:NEW_ASSETS_DIR
    The doc_path should be given in advance.
"""
def _get_hashed_file_name(file_path:str):
    file_path:pathlib.Path = pathlib.Path(file_path)
    try:
        with open(file_path,'r', encoding='utf-8') as file:
            file.read()
        file_hash = get_text_file_hash(file_path)
    except UnicodeDecodeError:
        file_hash = get_raw_file_hash(file_path)
    return file_hash+file_path.suffix

def _copy_figure(file_path):
    new_path = pathlib.Path(os.environ['NEW_ASSETS_DIR'])/_get_hashed_file_name(file_path)
    shutil.copy(file_path,new_path)
    return f"/{new_path.parent.name}/{new_path.name}" # add a root flag '/' to the new src's header

    
def _move_figure_to_new_place(elem:pf.Element,doc:pf.Doc,tracing_logger:TracingLogger,doc_path:pathlib.Path,**kwargs)->None:
    r"""Follow the general procedure of [Panflute](http://scorreia.com/software/panflute/)
    An `action` function to move local pictures to a new place (usually in a unified directory). Replace the original src with the new one.
    [modify elements in place]
    """
    if isinstance(elem, pf.Image) and (old_src:=str(elem.url)).startswith('.'): # reletive path
        old_src = decode_src_url(old_src)
        new_src = _copy_figure(doc_path.parent/old_src)
        tracing_logger.mark(elem)
        elem.url = new_src
        tracing_logger.check_and_log('image',elem)
    elif isinstance(elem, pf.RawInline) and elem.format == 'html' and (old_src:=get_html_src(elem.text)) and old_src.startswith('.'): # reletive path
        old_src = decode_src_url(old_src)
        new_src = _copy_figure(doc_path.parent/old_src)
        tracing_logger.mark(elem)
        elem.text = sub_html_src(elem.text,new_src)
        tracing_logger.check_and_log('raw_html_img',elem)

def run_filter(doc:pf.Doc=None,doc_path:pathlib.Path=None,**kwargs):
    assert doc_path.exists(),f"doc_path: {doc_path} does not exist."
    assert os.environ['NEW_ASSETS_DIR'], "NEW_ASSETS_DIR is not given in environment variables."
    return pf.run_filters(
        actions=[_move_figure_to_new_place],
        doc=doc,
        tracing_logger=TracingLogger(),
        doc_path=doc_path,**kwargs)


"""
See https://github.com/Zhaopudark/pandoc-filter for documentation.
"""
import logging

from . import filters

from .scripts import run_filters_pyio

from .version import __version__

from .utils.logging_helper import TracingLogger

logger = TracingLogger("./logs/pandoc_filter_log",level=logging.DEBUG)


import panflute as pf
from urllib.parse import urlparse

from ...utils.logging_helper import TracingLogger

r"""A pandoc filter that mainly for converting `markdown` to `html`.
Enhance the footnote.
"""
        
def _enhance_footnote(elem:pf.Element,doc:pf.Doc,tracing_logger:TracingLogger,**kwargs)->pf.Note|None:
    r"""Follow the general procedure of [Panflute](http://scorreia.com/software/panflute/)
    An action to process footnotes. Deal with the footnote content as follows:
        - upgrade the link-like string to a `link` element.
        - Remove unnecessary `\n`.
        - Remove the markdown link.
        - Normalize many markdown formats like emphasis(italic) and strong(bold).
    [replace elements]
    """
    if isinstance(elem, pf.Note):
        # elem = pf.Note(pf.Para(pf.Str(pf.stringify(elem).strip(' \n'))))
        # buf = 
        
        tracing_logger.mark(elem)
        buf = []
        for item in pf.stringify(elem).strip(' \n').split():
            
            url = urlparse(item)
            if (url.scheme and url.netloc):
                # tracing_logger.info('url:',url)
                # link = 
                buf.append(pf.Link(pf.Str(item),url=url.geturl()))
            elif url.path.lower().startswith('www'):
                # link = 
                buf.append(pf.Link(pf.Str(item),url='https://'+url.geturl()))
            else:
                buf.append(pf.Str(item))
            buf.append(pf.Space())
                
            # if isinstance(item,pf.Str):
            #     url = urlparse(item.text)
            #     if (url.scheme and url.netloc):
            #         # tracing_logger.info('url:',url)
            #         link = pf.Link(item,url=url.geturl())
            #         buf.append(link)
            #     elif url.path.lower().startswith('www'):
            #         link = pf.Link(item,url='https://'+url.geturl())
            #         buf.append(link)
            #     else:
            #         buf.append(item)
            # else:
            #     buf.append(item)
            #     # link = pf.Link(item,url=item.text)
            #     # item = link
        elem = pf.Note(pf.Para(*buf))
        
        tracing_logger.check_and_log('footnote',elem)
        return elem

    # if isinstance(elem, pf.Str):
    #     url  = urlparse(elem.text)
    #     if url.scheme and url.netloc:
    #         tracing_logger.info('url:',url)
            
        # tracing_logger.mark(elem)
        # link = pf.Link(elem,url=elem.text)
        
        # tracing_logger.check_and_log('link_like',elem)
        # return link

def run_filter(doc:pf.Doc=None,**kwargs):
    return pf.run_filters(actions=[_enhance_footnote],doc=doc,tracing_logger=TracingLogger(),**kwargs)


from . import convert_github_style_alert_to_hexo_style_alert
from . import enhance_equation
from . import norm_footnote
from . import norm_internal_link
from . import upload_figure_to_aliyun
from . import move_figure_to_new_place


import difflib
import pathlib
import subprocess
import functools
import pandoc_filter
import pytest

# import pandoc_filter.filters
# import pandoc_filter.filters.md2md
# import pandoc_filter.filters.md2md.norm_internal_link

def _check_file_path(file_path:str)->pathlib.Path:
    file_path:pathlib.Path = pathlib.Path(file_path)
    assert file_path.exists()
    assert file_path.is_file()
    return file_path

def _compare_files(file1_path, file2_path):
    with open(file1_path, 'r', encoding='utf-8') as file1:
        content1 = file1.readlines()

    with open(file2_path, 'r', encoding='utf-8') as file2:
        content2 = file2.readlines()

    differ = difflib.Differ()
    diff = list(differ.compare(content1, content2))
    if not any(line.startswith('- ') or line.startswith('+ ') for line in diff):
        return True
    else:
        return False
    
@pytest.mark.parametrize('input_format',['markdown'])
@pytest.mark.parametrize('output_format',['html'])
def test_md2html_header_anchor_link_filter(input_format:str,output_format:str):
    file_path = _check_file_path("./resources/inputs/test_md2html_header_anchor_and_link.md")
    pathlib.Path("./temp").mkdir(parents=True, exist_ok=True)
    output_path = pathlib.Path(f"./temp/{file_path.stem}.html")
    answer_path = pathlib.Path(f"./resources/outputs/{file_path.stem}.html")
    pandoc_command = [
        'pandoc',
        file_path,
        '-o',
        output_path,
        '-f',
        input_format,
        '-t',
        output_format,
        '-s',
        '--filter',
        'md2md-norm-internal-link-filter',
        '--filter',
        'md2html-hash-anchor-and-internal-link-filter',
        '--filter',
        'md2html-enhance-link-like-filter',
        '--filter',
        'md2html-increase-header-level-filter',
    ]
    assert subprocess.run(pandoc_command, check=True).returncode == 0
    assert _compare_files(output_path,answer_path)


@pytest.mark.parametrize('input_format',['markdown'])
@pytest.mark.parametrize('output_format',['html'])
def test_md2html_header_anchor_link_filter_pyio(input_format:str,output_format:str):
    file_path = _check_file_path("./resources/inputs/test_md2html_header_anchor_and_link.md")
    pathlib.Path("./temp").mkdir(parents=True, exist_ok=True)
    output_path = pathlib.Path(f"./temp/{file_path.stem}.html")
    answer_path = pathlib.Path(f"./resources/outputs/{file_path.stem}.html")
    pandoc_filter.run_filters_pyio(
        file_path,
        output_path,
        input_format,output_format,
        [pandoc_filter.filters.md2md.norm_internal_link.run_filter,
         pandoc_filter.filters.md2html.hash_anchor_and_internal_link.run_filter,
         pandoc_filter.filters.md2html.enhance_link_like.run_filter,
         pandoc_filter.filters.md2html.increase_header_level.run_filter])
    assert _compare_files(output_path,answer_path)

@pytest.mark.parametrize('input_format',['markdown'])
@pytest.mark.parametrize('output_format',['html'])
def test_md2html_enhance_footnote_filter_pyio(input_format:str,output_format:str):
    file_path = _check_file_path("./resources/inputs/test_md2html_footnote.md")
    pathlib.Path("./temp").mkdir(parents=True, exist_ok=True)
    output_path = pathlib.Path(f"./temp/{file_path.stem}.html")
    answer_path = pathlib.Path(f"./resources/outputs/{file_path.stem}.html")
    pandoc_filter.run_filters_pyio(
        file_path,
        output_path,
        input_format,output_format,
        [pandoc_filter.filters.md2html.enhance_footnote.run_filter,])
    assert _compare_files(output_path,answer_path)
   

@pytest.mark.parametrize('input_format',['markdown'])
@pytest.mark.parametrize('output_format',['html'])
def test_md2html_enhance_footnote_filter(input_format:str,output_format:str):
    file_path = _check_file_path("./resources/inputs/test_md2html_footnote.md")
    pathlib.Path("./temp").mkdir(parents=True, exist_ok=True)
    output_path = pathlib.Path(f"./temp/{file_path.stem}.html")
    answer_path = pathlib.Path(f"./resources/outputs/{file_path.stem}.html")
    pandoc_command = [
        'pandoc',
        file_path,
        '-o',
        output_path,
        '-f',
        input_format,
        '-t',
        output_format,
        '-s',
        '--filter',
        'md2html-enhance-footnote-filter',
    ]
    assert subprocess.run(pandoc_command, check=True).returncode == 0
    assert _compare_files(output_path,answer_path)




import panflute as pf

from ...utils.logging_helper import TracingLogger

r"""A pandoc filter that mainly for converting `markdown` to `markdown`.
Normalize the footnote.
"""
        
def _norm_footnote(elem:pf.Element,doc:pf.Doc,tracing_logger:TracingLogger,**kwargs)->pf.Note|None:
    r"""Follow the general procedure of [Panflute](http://scorreia.com/software/panflute/)
    An action to process footnote. Deal with the footnote content as follows:
        - Remove unnecessary `\n`.
        - Remove the markdown link.
        - Normalize many markdown formats like emphasis(italic) and strong(bold).
    [replace elements]
    """
    if isinstance(elem, pf.Note):
        tracing_logger.mark(elem)
        elem = pf.Note(pf.Para(pf.Str(pf.stringify(elem).strip(' \n'))))
        tracing_logger.check_and_log('footnote',elem)
        return elem

def run_filter(doc:pf.Doc=None,**kwargs):
    return pf.run_filters(actions=[_norm_footnote],doc=doc,tracing_logger=TracingLogger(),**kwargs)


import logging
import panflute as pf

from ...utils.logging_helper import TracingLogger

r"""A pandoc filter that mainly for converting `markdown` to `html`.
Increase the header level by 1.

In markdown, the header level is from 1 to 6, which is represented by `#` to `######`.
We may use `#` to represent the chapter, `##` to represent the section, and so on.
But in html, the first level header is `<h1>`, and the last level header is `<h6>`.
And, the first level header is usually used for the title of the whole document instead of the chapter.
So, we may want to increase the header level by 1 when converting markdown to html, to maintain a general html style structure.

"""

def _increase_header_level(elem:pf.Element,doc:pf.Doc,tracing_logger:TracingLogger,**kwargs)->None:
    r"""Follow the general procedure of [Panflute](http://scorreia.com/software/panflute/)
    An `action` function to increase the header level by 1.
    [modify elements in place]
    """
    if isinstance(elem, pf.Header):
        tracing_logger.mark(elem)
        if elem.level < 6:
            elem.level += 1
        else:
            elem.level = 6 #  truncate the level to 6
        tracing_logger.check_and_log('header',elem)
        

def run_filter(doc:pf.Doc=None,**kwargs):
    return pf.run_filters(actions= [_increase_header_level],doc=doc,tracing_logger=TracingLogger(),**kwargs)